#include <math.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <pthread.h>
#include <glib.h>
#include "patch.h"
#include "specimen.h"
#include "driver.h"
#include "sample.h"
#include "maths.h"
#include "ticks.h"

/* magic numbers */
enum {
     EVENTMAX = 1024,
};

/* event codes */
typedef enum {
     MIXER_NOTEON = 1,
     MIXER_NOTEOFF,
     MIXER_NOTEON_WITH_ID,
     MIXER_NOTEOFF_WITH_ID,
} MixerEvent;

/* type for ringbuffer of incoming events */
typedef struct event_s {
     MixerEvent type;		/* type of event */
     tick_t     ticks;		/* time of event's occurence */
     int        id;		/* patch id this applies to */
     int        chan;		/* channel */
     int        note;		/* midi note */
     int        vel;		/* midi velocity */
} event_t;

/* special structure for previewing samples */
typedef struct preview_s {
     int      active;
     sample_t sample;
     int      next_frame;
} preview_t;

/* general variables */
static float volume = 0.5;	/* master volume */
static preview_t preview;	/* current preview sample */
static event_t events[EVENTMAX];
static event_t *writer = events;
static event_t *reader = events;

/* advance reader event vector */
inline static void advance_reader()
{
     reader = (reader + 1 >= events + EVENTMAX) ?
	  events : reader +1;
}

/* advance writer event vector */
inline static void advance_writer()
{
     writer = (writer + 1 >= events + EVENTMAX) ? 
	  events : writer + 1;
}

/* mixdown current preview sample (if any) into buf */
inline static void preview_render(float *buf, int frames)
{
     register int i, j;

     /* if there is a sample currently being previewed let's throw it into the mix */
     if (preview.active && pthread_mutex_trylock(&preview.sample.mutex) == 0) {

     	  if (preview.sample.sp != NULL) {

	       /* fill the buf with samples from preview (two samples per frame) */
	       for (i = 0, j = preview.next_frame*2; i < frames*2 && j < preview.sample.frames*2; i++, j++)
		    buf[i] += preview.sample.sp[j];

	       /* increment next frame indicator (half as much as the number of samples written) */
	       if ((preview.next_frame = j/2) >= preview.sample.frames) {

		    preview.active = 0;

		    if (preview.sample.sp != NULL)
			 free(preview.sample.sp);

		    preview.sample.sp = NULL;
	       }
	  } else {
	       preview.active = 0;
	  }
	  pthread_mutex_unlock(&preview.sample.mutex);
     }
}

/* flush all events currently queued */
void mixer_flush()
{
     /* skip any queued ringbuffer events */
     reader = writer;
}

/* mix current soundscape into buf */
void mixer_mixdown(float *buf, int frames)
{
     tick_t curticks = getticks();
     int offset;
     register int i;

     /* clear the buffer */
     for (i = 0; i < frames*2; i++)
	  buf[i] = 0.0;
     
     /* get events */
     while (reader != writer) {
	  if (reader->ticks > curticks)
	       break;

	  offset = reader->ticks - (curticks - frames);
	  if (offset < 0)
	       offset = 0;
	  
	  switch (reader->type) {
	  case MIXER_NOTEON:
	       patch_trigger(reader->chan, reader->note, reader->vel / 127.0, reader->ticks, offset);
	       break;
	  case MIXER_NOTEON_WITH_ID:
	       patch_trigger_with_id(reader->id, reader->note, reader->vel / 127.0, reader->ticks, offset);
	       break;
	  case MIXER_NOTEOFF:
	       patch_release(reader->chan, reader->note, offset);
	       break;
	  case MIXER_NOTEOFF_WITH_ID:
	       patch_release_with_id(reader->id, reader->note, offset);
	       break;
	  }

	  advance_reader();
     }

     /* render data */
     patch_render(buf, frames);
     preview_render(buf, frames);

     /* scale to master volume */
     for (i = 0; i < frames*2; i++)
	  buf[i] *= volume;
}

/* queue a note-off event */
void mixer_note_off(int chan, int note)
{
     writer->type = MIXER_NOTEOFF;
     writer->ticks = getticks();
     writer->chan = chan;
     writer->note = note;
     advance_writer();
}

/* queue a note-off event by patch id */
void mixer_note_off_with_id(int id, int note)
{
     writer->type = MIXER_NOTEOFF_WITH_ID;
     writer->ticks = getticks();
     writer->id = id;
     writer->note = note;
     advance_writer();
}

/* queue a note-on event */
void mixer_note_on(int chan, int note, int vel)
{
     writer->type = MIXER_NOTEON;
     writer->ticks = getticks();
     writer->chan = chan;
     writer->note = note;
     writer->vel = vel;
     advance_writer();
}

/* queue a note-on event by patch id */
void mixer_note_on_with_id(int id, int note, int vel)
{
     writer->type = MIXER_NOTEON_WITH_ID;
     writer->ticks = getticks();
     writer->id = id;
     writer->note = note;
     writer->vel = vel;
     advance_writer();
}

/* preview a sample file */
void mixer_preview(char *name)
{
     preview.active = 0;
     sample_load_file(&preview.sample, name);
     preview.next_frame = 0;
     preview.active = 1;
}

/* set the master volume */
int mixer_set_volume(float vol)
{
     if (vol < 0.0 || vol > 1.0)
	  return -1;

     volume = vol;
     return 0;
}
